/*--------------------------------------------------------------+
| Author: Daniel Porumbel 2018                                  |
|License: Creative Commons Attribution 3.0 Unported License     |
|         http://creativecommons.org/licenses/by/3.0/           |             
+--------------------------------------------------------------*/
#include <iostream>
#include "bendersSubprob.h"
#include <list>
#include <vector>
#include <ilcplex/ilocplex.h>                    //Cplex header
using namespace std;
#define THREADS 1

double*b;                                        //the demands

//#define PRINT_OPTSOL_OF_BENDERS_POLYTOPE
int isEdge(int v1, int v2, double * y){
    if(y==NULL)
        return g[v1][v2];
    if(g[v1][v2])
        if(y[gEdge[v1][v2]]>EPSILON)
            return 1;
    return 0;
}
void fillEdgesBalancedBFS(double *edgesTree, int source){
    int** tree = new int*[n];
    int* connDepth = new int[n];
    for(int i=0;i<n;i++){
        tree[i] = new int[n];
        connDepth[i] = 0;//depth 0: not yet discovered
        for(int j=0;j<n;j++)
            tree[i][j] = 0;
    }
    list<int> queue;
    queue.push_back(source); 
    connDepth[source] = 1;

    while(!queue.empty()){
        int currVtx   = queue.front();
        queue.pop_front();
        for(int v=0;v<n;v++)
            if(v!=currVtx)
                if(g[currVtx][v])
                    if(connDepth[v]==0){//not yet discovered
                        connDepth[v] = connDepth[currVtx]+1;
                        queue.push_back(v);
                        vector<int> parents;
                        int nbParents = 0;
                        for(int p=0;p<n;p++)
                            if((g[p][v])&&(p!=v)){
                               if(connDepth[p]==connDepth[currVtx]){
                                    parents.push_back(p); 
                                    nbParents++;
                               }
                            }
                        assert(nbParents>0);
                        //choosing random parents of depth connDepth[currVtx];
                        int rndParent = parents[rand()%nbParents];
                        assert(rndParent>=0);
                        assert(rndParent<n);
                        tree[rndParent][v]=1;
                        tree[v][rndParent]=1;
                    }
    }
    int nrEdges = 0;
    for(int i=n-1;i>=0;i--){
        clog<<i<<":";
        for(int j=0;j<n;j++)
            if(tree[i][j]){
                nrEdges++;
                clog<<j<<"("<<connDepth[j]<<") ";
            }
        clog<<endl;
    }
    //filling the edges
    for(int e=0;e<m;e++){
        edgesTree[e]= 0;
        int vtx1 = ii[e];
        int vtx2 = jj[e];
        if(tree[vtx1][vtx2])
            edgesTree[e] = 1;
    }
    if(!(nrEdges==2*(n-1))){
        cerr<<"Can not find a spanning tree covering everything. Is the graph connected?\n";
    }
    for(int i=0;i<n;i++)
        if(!connDepth[i]){
            cerr<<"Vtx "<<i<<" not connected. Strange\n";
        }
    delete[] connDepth;
    for(int i=0;i<n;i++)
        delete[] tree[i];
    delete[] tree;
}
void fillConnectedUpToDepth_BFS(int* connected, double* y, int source, int depth){
    list<int> queue;
    list<int> queueDepth;
    queue.push_back(source); 
    queueDepth.push_back(depth); 
    connected[source] = 1;

    while(!queue.empty()){
        int currentVtx   = queue.front();
        queue.pop_front();
        int depthCurrVtx = queueDepth.front();
        queueDepth.pop_front();
        if(depthCurrVtx>0)
          for(int v=0;v<n;v++)
            if(v!=currentVtx)
                if(isEdge(currentVtx,v,y))
                    if(!connected[v]){
                        connected[v] = 1;
                        queue.push_back(v);
                        queueDepth.push_back(depthCurrVtx-1);
                    }
    }
}

void fillConnectedUpToDepth(int* connected, double* y, int currentVtx, int depth){
    fillConnectedUpToDepth_BFS(connected,y,currentVtx,depth);
    for(int i=0;i<n;i++)
        if(connected[i]>0)
            connected[i] = 1;
}
int newCutSet(int* connected, double *cutEdges, double & bSinks){
    bSinks = 0;
    for(int v=0;v<n;v++)
        if(!connected[v])
            bSinks+=b[v];
    for(int e=0;e<m;e++)
        cutEdges[e] = 0;
    for(int e=0;e<m;e++){
        int vtx1 = ii[e];
        int vtx2 = jj[e];
        if(connected[vtx2]){
            int tmp = vtx1;
            vtx1    = vtx2;
            vtx2    = tmp;
        }
        //if there is any connected, then vtx1 is connected
        if(connected[vtx1]){
            if(!connected[vtx2]){
                cutEdges[e] = 1;
            }
        }
    }//end scan edges
    return (bSinks>0);
}

//classical separation cut
//Input yyy of size m
//Output u of size m
void newCut(double * yyy, double *u, double &rightHand){
    int i;
    IloEnv env;
    IloModel model(env);
    IloNumVarArray uu(env, m+n-1, 0, 10000000);
    
    IloExpr expr(env);
    for(i=0;i<m;i++)
        expr-=(yyy[i])*uu[i];
    for(i=m;i<m+n-1;i++){
        expr+=(b[i-m])*uu[i];
        //clog<<"b["<<i<<"-"<<m<<"]="<<b[i-m]<<endl;
    }
    IloObjective obj(env, expr, IloObjective::Maximize);
    model.add(obj);

    IloRangeArray constr(env);
    IloExpr e(env);
    for(i=0;i<m+n-1;i++)
        e += uu[i];
        //e += (i+1)*uu[i];
    constr.add(e==1);

    for(i=0;i<m;i++){
        int vtx1 = ii[i];
        int vtx2 = jj[i];
        //clog<<"vtx1="<<vtx1<<","; clog<<"vtx2="<<vtx2<<endl;
        if((vtx1!=n-1)&&(vtx2!=n-1)){
            IloExpr e1(env);
            e1+=uu[i]-uu[vtx1+m]+uu[vtx2+m];
            constr.add(e1>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e1.end();

            IloExpr e2(env);
            e2+=uu[i]-uu[vtx2+m]+uu[vtx1+m];
            constr.add(e2>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e2.end();
        }
        if(vtx1==n-1){
            IloExpr e1(env);
            e1+=uu[i]+uu[vtx2+m];
            constr.add(e1>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e1.end();

            IloExpr e2(env);
            e2+=uu[i]-uu[vtx2+m];
            constr.add(e2>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e2.end();
        }
        if(vtx2==n-1){
            IloExpr e1(env);
            e1+=uu[i]-uu[vtx1+m];
            constr.add(e1>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e1.end();

            IloExpr e2(env);
            e2+=uu[i]+uu[vtx1+m];
            constr.add(e2>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e2.end();
        }
    }
    model.add(constr);
    //Solve model
    IloCplex cplex (env);
    cplex.extract(model);
    cplex.setOut(env.getNullStream());
    cplex.setError(env.getNullStream());
    cplex.setParam(IloCplex::Threads, THREADS);
    cplex.setParam(IloCplex::NumericalEmphasis, true);
    cplex.solve();

    //Extract solution
    double objVal = cplex.getObjValue();
    clog<<"   Classical separation cut to be added with objVal="<<objVal<<endl;
    IloNumArray solution(env);
    cplex.getValues (solution,uu);
    for(i=0;i<m;i++){
            u[i] = solution[i];
            #ifdef PRINT_OPTSOL_OF_BENDERS_POLYTOPE
            clog<<"u["<<i<<"]="<<u[i]<<",";         
            #endif
    }
    rightHand = 0;
    for(i=m;i<m+n-1;i++){
        #ifdef PRINT_OPTSOL_OF_BENDERS_POLYTOPE
        clog<<"u["<<i<<"]="<<solution[i]<<",";         
        #endif
        rightHand += b[i-m]*solution[i];
    }
    clog<<endl;
    clog<<"    generated separation constraint has rightHand="<<rightHand<<endl;
    //cplex.exportModel("test.lp");

    //Clean all and return
    constr.end();
    uu.end();
    model.end();
    env.end();
}

void newCutCheckRayHitsP(double * yyy, double *u, double &rightHand,double &tStar){
    int i;
    IloEnv env;
    IloModel model(env);
    IloNumVarArray uu(env, m+n-1, 0, 10000000);
    
    IloExpr expr(env);
    for(i=m;i<m+n-1;i++){
        expr+=(b[i-m])*uu[i];
        //clog<<"b["<<i<<"-"<<m<<"]="<<b[i-m]<<endl;
    }
    IloObjective obj(env, expr, IloObjective::Maximize);
    model.add(obj);
    
    IloRangeArray constr(env);

    IloExpr e(env);
    for(i=0;i<m+n-1;i++)
        e += (i+1)* uu[i];
    constr.add(e==20);
    
    IloExpr e2(env);
    for(i=0;i<m;i++)
        e2 += yyy[i]* uu[i];
    constr.add(e2==0);
    


    for(i=0;i<m;i++){
        int vtx1 = ii[i];
        int vtx2 = jj[i];
        //clog<<"vtx1="<<vtx1<<","; clog<<"vtx2="<<vtx2<<endl;
        if((vtx1!=n-1)&&(vtx2!=n-1)){
            IloExpr e1(env);
            e1+=uu[i]-uu[vtx1+m]+uu[vtx2+m];
            constr.add(e1>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e1.end();

            IloExpr e2(env);
            e2+=uu[i]-uu[vtx2+m]+uu[vtx1+m];
            constr.add(e2>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e2.end();
        }
        if(vtx1==n-1){
            IloExpr e1(env);
            e1+=uu[i]+uu[vtx2+m];
            constr.add(e1>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e1.end();

            IloExpr e2(env);
            e2+=uu[i]-uu[vtx2+m];
            constr.add(e2>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e2.end();
        }
        if(vtx2==n-1){
            IloExpr e1(env);
            e1+=uu[i]-uu[vtx1+m];
            constr.add(e1>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e1.end();

            IloExpr e2(env);
            e2+=uu[i]+uu[vtx1+m];
            constr.add(e2>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e2.end();
        }
    }
    model.add(constr);
    //Solve model
    IloCplex cplex (env);
    cplex.extract(model);
    cplex.setOut(env.getNullStream());
    cplex.setError(env.getNullStream());
    //This gives Markowitz threshold set to 0.1
    cplex.setWarning(env.getNullStream());
    cplex.setParam(IloCplex::Threads, THREADS);
    cplex.setParam(IloCplex::NumericalEmphasis, true);
    cplex.solve();

    //Extract solution
    double objVal;
    try{
        objVal = cplex.getObjValue();
    }catch (IloCplex::Exception e){
       cerr<<"Concert Exception Ray Feasibility:"<<e.getMessage()<<endl;
       cerr<<"Model saved to trycatch.lp. I. I think there is no problem now";
       cplex.exportModel("trycatch.lp");
       tStar = 0;
       return ;
    }
    clog<<"   Testing the polytope is hit by ray, case (1), and got objVal="<<objVal<<endl;
    if(objVal>EPSILON)
        tStar = INFEASIBLE;
    else
        tStar = 0;
    IloNumArray solution(env);
    cplex.getValues (solution,uu);
    for(i=0;i<m;i++){
            u[i] = solution[i];
            #ifdef PRINT_OPTSOL_OF_BENDERS_POLYTOPE
            clog<<"u["<<i<<"]="<<u[i]<<",";         
            #endif
    }
    if(tStar == INFEASIBLE){
        clog<<"   Constraint separating whole ray: ";
    }
        rightHand = 0;
        for(i=m;i<m+n-1;i++){
            #ifdef PRINT_OPTSOL_OF_BENDERS_POLYTOPE
            clog<<"u["<<i<<"]="<<solution[i]<<",";         
            #endif
            rightHand += b[i-m]*solution[i];
        }
        assert(max(objVal-rightHand,rightHand-objVal)<EPSILON*10);
        clog<<"   Intersect algo for infeasible case rrightHand="<<rightHand<<endl;
    //cplex.exportModel("test.lp");

    //Clean all and return
    constr.end();
    uu.end();
    model.end();
    env.end();

}
void newIntersectionFeasibleCase(double * yyy, double *u, double &rightHand,double &tStar){ 
    clog<<"      bendersSubprob:Start Feasible Intersection"<<endl;
    int i;
    IloEnv env;
    IloModel model(env);
    IloNumVarArray uuu(env, m+n-1, 0, 10000000);
    
    IloExpr expr(env);
    clog<<"      bendersSubprob: objective func is (in vars uuu)";
    for(i=m;i<m+n-1;i++){
        expr+=(b[i-m])*uuu[i];
        clog<<"b["<<i-m<<"]="<<b[i-m]<<"*uuu["<<i<<"]+";
    }
    clog<<endl;
    IloObjective obj(env, expr, IloObjective::Maximize);
    model.add(obj);

    IloRangeArray constr(env);
    IloExpr e(env);
    for(i=0;i<m;i++)
        if(yyy[i]>EPSILON)
            e += yyy[i]*uuu[i];
    constr.add(e==1);

    for(i=0;i<m;i++){
        int vtx1 = ii[i];
        int vtx2 = jj[i];
        //clog<<"vtx1="<<vtx1<<","; clog<<"vtx2="<<vtx2<<endl;
        if((vtx1!=n-1)&&(vtx2!=n-1)){
            IloExpr e1(env);
            e1+=uuu[i]-uuu[vtx1+m]+uuu[vtx2+m];
            constr.add(e1>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e1.end();

            IloExpr e2(env);
            e2+=uuu[i]-uuu[vtx2+m]+uuu[vtx1+m];
            constr.add(e2>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e2.end();
        }
        if(vtx1==n-1){
            IloExpr e1(env);
            e1+=uuu[i]+uuu[vtx2+m];
            constr.add(e1>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e1.end();

            IloExpr e2(env);
            e2+=uuu[i]-uuu[vtx2+m];
            constr.add(e2>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e2.end();
        }
        if(vtx2==n-1){
            IloExpr e1(env);
            e1+=uuu[i]-uuu[vtx1+m];
            constr.add(e1>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e1.end();

            IloExpr e2(env);
            e2+=uuu[i]+uuu[vtx1+m];
            constr.add(e2>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e2.end();
        }
    }
    model.add(constr);
    //Solve model
    IloCplex cplex (env);
    cplex.extract(model);
    cplex.setOut(env.getNullStream());
    cplex.setError(env.getNullStream());
    cplex.setWarning(env.getNullStream());
    cplex.setParam(IloCplex::Threads, THREADS);
    cplex.setParam(IloCplex::NumericalEmphasis, true);
    double objVal = 0;
    IloNumArray solution(env);
    try{
        cplex.solve();
        objVal = cplex.getObjValue();
        clog<<"Objval in classical intersection="<<objVal<<endl;
        //Extract solution
        cplex.getValues (solution,uuu);
    }catch(IloCplex::Exception e){
       cplex.exportModel("bendersProb.lp");
       cerr<<"\n\n\nError in intersection subproblem"<<e.getMessage()<<endl;
       cerr<<"Model saved to bendersProb.lp. I. I think the problem is infeasible\n\
              (maybe there is no sink)\n\
              (maybe a vertex v with b_v>0 is not connected to any edge). Return \n";
       cerr<<"I do not know the state of the returned arguments.\n\n\n";
       return;
    }
    catch(IloAlgorithm::NotExtractedException e){
       cplex.exportModel("bendersProbNotExtracted.lp");
       cerr<<"\n\n\nError in intersection subproblem"<<e.getMessage()<<endl;
       cerr<<"Model saved to bendersProbNotExtracted.lp. I. I think the problem is infeasible\n\
              (maybe there is no sink)\n\
              (maybe a vertex v with b_v>0 is not connected to any edge). Return \n";
       cerr<<"NotExtractedException was very rare too me.\n\n\n";
       for(i=m;i<m+n-1;i++){
            //clog<<"getting "<<i<<" again,";
            try{
                cplex.getValue (uuu[i]);
                clog<<"extracting "<<i<<",";
            }catch(IloAlgorithm::NotExtractedException){
                clog<<"I think uuu["<<i<<"] does not arises in model\n";
                if(i>=m){
                    clog<<i<<"corresponds to vtx"<<i-m<<endl;
                    clog<<"b["<<i-m<<"]="<<b[i-m]<<endl;
                    clog<<"Line of adjacency matrix for "<<i-m<<endl;
                    int countNeighbors = 0;
                    for(int j=0;j<n;j++){
                        clog<<g[i-m][j];
                        if(g[i-m][j]==1)
                            countNeighbors++;
                    }
                    clog<<endl;
                    if((countNeighbors==0)&&(b[i-m]==0)){
                        clog<<"This vertex has no demand and it is isolated.  Forcing zero on it and re-optimizing.\n";
                        //constr.add(uuu[i]==0);
                        model.add(uuu[i]==0);
                        cplex.solve();
                        clog<<"Resolved "<<i<<". I will extract it again\n";
                        i--;
                    }
                }
            }
       }
       clog<<"Trying to extract solution again:\n";
       objVal = cplex.getObjValue();
       clog<<"New objVal="<<objVal<<endl;
       //Extract solution
       //for(i=m;i<m+n-1;i++){
       //     try{
       //         cplex.getValue (uuu[i]);
       //     }catch(IloAlgorithm::NotExtractedException){
       //         clog<<"I think uuu["<<i<<"] does not arises in model\n";
       //         exit(1);
       //     }
       //}
       cplex.getValues (solution,uuu);
    }
    clog<<"    Intersection algo solved classical hit intersection and obtained      tStar="<<objVal<<endl;
    tStar = objVal;
    for(i=0;i<m;i++){
            u[i] = solution[i];
            //u[i] is a free variable in our model, so we assign it the smallest
            //feasible value
            if(u[i]==10000000) {
                if((ii[i]!=n-1)&&(jj[i]!=n-1)){
                     u[i] = max(solution[m+ii[i]]-solution[m+jj[i]],
                           -solution[m+ii[i]]+solution[m+jj[i]]);
                }else{
                    if(ii[i]==n-1)
                      u[i] = max(-solution[m+jj[i]],+solution[m+jj[i]]);
                    if(jj[i]==n-1)
                      u[i] = max(-solution[m+ii[i]],+solution[m+ii[i]]);
                }
            }
            #ifdef PRINT_OPTSOL_OF_BENDERS_POLYTOPE
            clog<<"uuu["<<i<<"]="<<u[i]<<"/";         
            clog<<"yyy["<<i<<"]="<<yyy[i]<<",";
            #endif
    }
    rightHand = 0;
    for(i=m;i<m+n-1;i++){
        #ifdef PRINT_OPTSOL_OF_BENDERS_POLYTOPE
        clog<<"uuu["<<i<<"]="<<solution[i]<<",";         
        #endif
        rightHand += b[i-m]*solution[i];
    }
    clog<<"    Intersect Algo (feasible case) determined constraint with rrrightHand="<<rightHand<<endl;
    //cplex.exportModel("test.lp");

    //Clean all and return
    constr.end();
    uuu.end();
    model.end();
    env.end();

    //report upper bound
    if(tStar>100000){
        clog<<"Warning: huge feasible tStar="<<tStar<<" Set to INFEASIBLE\n";
        tStar = INFEASIBLE;
    }
}
