/*--------------------------------------------------------------+
| Author: Daniel Porumbel 2018                                  |
|License: Creative Commons Attribution 3.0 Unported License     |
|         http://creativecommons.org/licenses/by/3.0/           |             
+--------------------------------------------------------------*/
#ifndef GENERAL_H_INCLUDED
#define GENERAL_H_INCLUDED
#include <iomanip>
#include <locale>
#include <sstream>

/*
 * =================================================================================================
 *                  VARIOUS GENERAL ROUTINES (eg., getCPUTime() due to  D.R. Nadeau)
 * =================================================================================================
 */
extern "C" {
  #if defined(_WIN32)
  #include <Windows.h>
  
  #elif defined(__unix__) || defined(__unix) || defined(unix) || (defined(__APPLE__) && defined(__MACH__))
  #include <unistd.h>
  #include <sys/resource.h>
  #include <sys/times.h>
  #include <time.h>
  
  #else
  #error "Unable to define getCPUTime( ) for an unknown OS."
  #endif
  
  /**
   * Returns the amount of CPU time (including parallel usage) used by the current process,
   * in *fractional* seconds, or -1.0 if an error occurred.
   */
  double getCPUTime( );
}

std::string toString(int number);

#endif
